#include <stdio.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netdb.h>

#ifndef lint
static char Version[] = "@(#)rping.c	e07@nikhef.nl (Eric Wassenaar) 910926";
#endif lint

extern int errno;
extern char *rindex();
extern char *strcpy();
extern char *inet_ntoa();

u_short checksum();
char *pr_type();
char *pr_addr();

#ifdef IP_OPTIONS
#ifndef MAX_IPOPTLEN
#define MAX_IPOPTLEN 40
#endif
#endif IP_OPTIONS

#define NOT_DOTTED_QUAD ((unsigned)-1)

#define TRUE	1
#define FALSE	0

#define MAXADDRS 35
#define MAXDNAME 256

#define	MAXPKT	4096		/* max packet size */
#define PKTSIZE	64		/* icmp packet total size */
#define HDRLEN	ICMP_MINLEN	/* icmp header length */
#define DATALEN	PKTSIZE-HDRLEN	/* icmp packet data length */

u_char packet[MAXPKT];		/* packet buffer */

int attempts = 3;		/* number of retries */
int timeout = 2;		/* timeout in secs */
int verbose = 0;		/* verbose output */
int traceroute = FALSE;		/* gather routing info */

long timelimit;			/* safety limit */
char *program;			/* program name */

#define EX_ALIVE 0		/* host is alive */
#define EX_DEAD  1		/* host is dead */
#define EX_USAGE 2		/* illegal parameter */
#define EX_ERROR 3		/* system error */

char Usage[] = "Usage: %s [-t secs] [-r retries] [-R] [-v[v]] host\n";


main(argc, argv)
int argc;
char *argv[];
{
	int status;
	register char *opt;

	setlinebuf(stdout);

	if ((program = rindex(argv[0], '/')) != NULL)
		program++;
	else
		program = argv[0];

	while (argc > 2 && argv[1][0] == '-')
	{
	    for (opt = &argv[1][1]; *opt; opt++)
	    {
	    	switch (*opt)
	    	{
	    	    case 'r':
			--argc; argv++;
			attempts = atoi(argv[1]);
	    		break;

	    	    case 't':
			--argc; argv++;
			timeout = atoi(argv[1]);
	    		break;

	    	    case 'v':
			verbose++;
			break;

	    	    case 'R':
			traceroute = TRUE;
	    		break;

	    	    default:
    			fprintf(stderr, Usage, program);
	    		exit(EX_USAGE);
	    	}
	    }

	    --argc; argv++;
	}

	if (argc != 2 || argv[1][0] == '-')
	{
    		fprintf(stderr, Usage, program);
    		exit(EX_USAGE);
	}

	timelimit = time(0) + (attempts * timeout);

#ifdef OMNINET
	if (traceroute)
		(void) initdevice(OMNINET);
#endif OMNINET

	status = ping(argv[1]);
	exit(status);
}


ping(host)
char *host;				/* name of host to be pinged */
{
	struct hostent *hp;
	char hostname[MAXDNAME+1];
	struct in_addr inaddr[MAXADDRS];
	u_long addr;
	int naddrs;
	int status;
	register int i;

/*
 * Determine remote host address(es).
 */
	if (host == NULL)
		return(EX_ERROR);

	addr = inet_addr(host);
	if (addr != NOT_DOTTED_QUAD)
	{
		hp = gethostbyaddr((char *)&addr, sizeof(addr), AF_INET);
		if (hp != NULL)
			host = strcpy(hostname, hp->h_name);

		inaddr[0].s_addr = addr;
		naddrs = 1;
	}
	else
	{
		hp = gethostbyname(host);
		if (hp == NULL)
		{
			error("Unknown host: %s", host);
			return(EX_ERROR);
		}
		host = strcpy(hostname, hp->h_name);

		for (i = 0; i < MAXADDRS && hp->h_addr_list[i]; i++)
			inaddr[i] = *(struct in_addr *)hp->h_addr_list[i];
		naddrs = i;
	}

/*
 * Send probe packet to each address of the remote host.
 */
	for (i = 0; i < naddrs; i++)
	{
		if (verbose > 1 && naddrs > 1)
			printf("Trying %s ...\n", inet_ntoa(inaddr[i]));

		status = ping_host(host, inaddr[i]);
		if (status == EX_ERROR)
			break;

		if (status == EX_ALIVE && verbose < 2)
			break;
	}

	return(status);
}


ping_host(host, inaddr)
char *host;				/* name of host to be pinged */
struct in_addr inaddr;			/* network address of host */
{
	struct protoent *proto;
	struct sockaddr toaddr, fromaddr;
	struct sockaddr_in *to   = (struct sockaddr_in *)&toaddr;
	struct sockaddr_in *from = (struct sockaddr_in *)&fromaddr;
	int addrlen;
	int sock;
	int status;

	struct icmp *icmp = (struct icmp *)packet;
	struct timeval *timep = (struct timeval *)&packet[HDRLEN];
	u_char *datap = &packet[HDRLEN+sizeof(struct timeval)];

	int ident = getpid() & 0xFFFF;	/* ident of icmp output packet */
	int sequence = 0;		/* sequence of icmp output packet */
	int cc = PKTSIZE;		/* size of icmp output packet */
	int nc;				/* total size of input packet */

/*
 * Set network address.
 */
	to->sin_family = AF_INET;
	to->sin_addr.s_addr = inaddr.s_addr;
	to->sin_port = 0;

/*
 * Allocate socket.
 */
	proto = getprotobyname("icmp");
	if (proto == NULL)
	{
		error("Unknown protocol: icmp");
		return(EX_ERROR);
	}

	sock = socket(AF_INET, SOCK_RAW, proto->p_proto);
	if (sock < 0)
	{
		perror("socket");
		return(EX_ERROR);
	}

#ifdef IP_OPTIONS
	if (traceroute)
	{
		char optval[MAX_IPOPTLEN];

		bzero(optval, sizeof(optval));
		optval[IPOPT_OPTVAL]   = (char)IPOPT_RR;
		optval[IPOPT_OLEN]     = (char)MAX_IPOPTLEN - 1,
		optval[IPOPT_OFFSET]   = (char)IPOPT_MINOFF;
		optval[MAX_IPOPTLEN-1] = (char)IPOPT_EOL;

		proto = getprotobyname("ip");
		if (proto == NULL)
		{
			error("Unknown protocol: ip");
			return(EX_ERROR);
		}

		if (setsockopt(sock, proto->p_proto, IP_OPTIONS, optval, MAX_IPOPTLEN) < 0)
		{
			perror("setsockopt");
			return(EX_ERROR);
		}
	}
#endif IP_OPTIONS

	if (verbose)
		printf("[%s] ", host), (void) fflush(stdout);

/*
 * Prepare output packet.
 * Store current time in data field after packet header.
 */
resend:
	timep = (struct timeval *)&packet[HDRLEN];
	(void) gettimeofday(timep, (struct timezone *)NULL);

	datap = &packet[HDRLEN+sizeof(struct timeval)];
	for (nc = sizeof(struct timeval); nc < DATALEN; nc++)
		*datap++ = nc;

	icmp = (struct icmp *)packet;
	icmp->icmp_type  = ICMP_ECHO;
	icmp->icmp_code  = 0;
	icmp->icmp_cksum = 0;
	icmp->icmp_id    = ident;
	icmp->icmp_seq   = sequence;
	icmp->icmp_cksum = checksum((u_short *)icmp, cc);

/*
 * Send probe packet to remote destination.
 */
	nc = sendto(sock, (char *)packet, cc, 0, &toaddr, sizeof(toaddr));
	if (nc < 0)
	{
		perror("sendto");
		(void) close(sock);
		return(EX_ERROR);
	}

/*
 * Wait until input packet arrives. Retry after timeout.
 */
reread:
	status = wait_sock(sock, timeout);
	if (status < 0)
	{
		perror("select");
		(void) close(sock);
		return(EX_ERROR);
	}
	else if (status == 0)
	{
		if (++sequence < attempts)
			goto resend;

		errno = ETIMEDOUT;
		if (verbose)
			perror("recvfrom");
		status = EX_DEAD;
		goto done;
	}

/*
 * Fetch available input packet.
 * Note. On Apollo, recvfrom() will not be interrupted by an alarm().
 */
	addrlen = sizeof(fromaddr);
	nc = recvfrom(sock, (char *)packet, sizeof(packet), 0, &fromaddr, &addrlen);
	if (nc <= 0)
	{
		if (nc == 0)
			errno = ECONNRESET;

		if (errno == EINTR)
		{
			if (++sequence < attempts)
				goto resend;

			errno = ETIMEDOUT;
		}

		perror("recvfrom");
		(void) close(sock);
		return(EX_ERROR);
	}

/*
 * Analyze input packet. Reread if this is not what we want.
 */
	status = check_packet(packet, nc, cc, ident, sequence, from);
	if (status != EX_ALIVE)
	{
		if (time(0) <= timelimit)
			goto reread;
	}

done:
	if (!verbose)
		printf("%s is %s\n",
			host, (status == EX_ALIVE ? "alive" : "dead"));

	(void) close(sock);
	return(status);
}


wait_sock(sock, secs)
int sock;
int secs;
{
	fd_set fds;
	struct timeval wait;
	int nreply;

	wait.tv_sec = secs;
	wait.tv_usec = 0;

	/* FD_ZERO(&fds); */
	bzero((char *)&fds, sizeof(fds));
	FD_SET(sock, &fds);

	nreply = select(FD_SETSIZE, &fds, (fd_set *)0, (fd_set *)0, &wait);
	return(nreply);
}


check_packet(buf, nc, cc, ident, sequence, from)
u_char *buf;				/* address of input packet */
int nc;					/* total size of input packet */
int cc;					/* size of icmp output packet */
int ident;				/* ident in icmp output packet */
int sequence;				/* sequence in icmp output packet */
struct sockaddr_in *from;		/* address we got input packet from */
{
	struct ip *ip;
	struct icmp *icmp;
	int iphdrlen;			/* ip header length of input packet */
	struct timeval *sendtime;	/* time as stored in icmp packet */
	struct timeval recvtime;	/* time input packet is received */
	int triptime;			/* round trip time in millisecs */
	char *host;			/* name of host we got packet from */

	(void) gettimeofday(&recvtime, (struct timezone *)NULL);

	/* from->sin_addr.s_addr = ntohl(from->sin_addr.s_addr); */
	host = inet_ntoa(from->sin_addr);

/*
 * The input packet contains the ip header in front of the icmp packet.
 * Make sure the packet contains the icmp header after the ip header.
 */
	ip = (struct ip *)buf;
	iphdrlen = ip->ip_hl << 2;

	if (nc < iphdrlen + ICMP_MINLEN)
	{
		if (verbose)
			printf("%d bytes from %s (packet too short)\n", nc, host);
		return(EX_DEAD);
	}

	nc -= iphdrlen;
	icmp = (struct icmp *)(buf + iphdrlen);

	if (verbose)
		printf("%d bytes from %s ", nc, host);

/*
 * Check if this is really an echo reply packet.
 */
	if (icmp->icmp_type != ICMP_ECHOREPLY)
	{
		if (verbose)
			printf("(not echo reply: %s)\n", pr_type(icmp->icmp_type, icmp->icmp_code));
		return(EX_DEAD);
	}

/*
 * Check the ident of the icmp packet.
 */
	if (icmp->icmp_id != ident)
	{
		if (verbose)
			printf("(not from us: ident %d should be %d)\n", icmp->icmp_id, ident);
		return(EX_DEAD);
	}

/*
 * Check the sequence of the icmp packet.
 */
	if (icmp->icmp_seq != sequence)
	{
		if (verbose)
			printf("(out of sequence: seq %d should be %d)\n", icmp->icmp_seq, sequence);
		return(EX_DEAD);
	}

/*
 * Check the size of the icmp packet.
 */
	if (nc != cc)
	{
		if (verbose)
			printf("(packet size error: size %d should be %d)\n", nc, cc);
		return(EX_DEAD);
	}

/*
 * This seems to be our reply packet. Compute round trip time.
 */
	sendtime = (struct timeval *)&icmp->icmp_data[0];
	if ((recvtime.tv_usec -= sendtime->tv_usec) < 0)
	{
		recvtime.tv_sec--;
		recvtime.tv_usec += 1000000;
	}
	recvtime.tv_sec -= sendtime->tv_sec;
	triptime = (recvtime.tv_sec*1000) + (recvtime.tv_usec/1000);

	if (verbose)
		printf("seq=%d ttl=%d time=%d ms\n",
			sequence, ip->ip_ttl, triptime);

/*
 * The options buffer may be present at the end of the ip header.
 */
#ifdef IP_OPTIONS
	if (traceroute && (iphdrlen > sizeof(struct ip) + IPOPT_MINOFF-1))
	{
		char *ipopt;		/* address of options buffer */
		int ipoptlen;		/* total size of options buffer */
		int optlen;		/* length of options data */
		char *optptr;		/* pointer to options data */

		ipopt = (char *)(buf + sizeof(struct ip));
		ipoptlen = ipopt[IPOPT_OLEN];
		optlen = ipopt[IPOPT_OFFSET];
		optptr = &ipopt[IPOPT_MINOFF-1];

		if (ipopt[IPOPT_OPTVAL] == IPOPT_RR)
		{
			register int opt;

			for (opt = 1; optlen > IPOPT_MINOFF; opt++)
			{
				struct in_addr inaddr;
				bcopy(optptr, (char *)&inaddr, sizeof(inaddr));
				printf("%d  %s\n", opt, pr_addr(inaddr));
				optptr += sizeof(inaddr);
				optlen -= sizeof(inaddr);
			}
		}
	}
#endif IP_OPTIONS

	return(EX_ALIVE);
}


char *
pr_addr(inaddr)
struct in_addr inaddr;
{
	static char buf[BUFSIZ];
	struct hostent *hp;
	u_long addr;

	addr = inaddr.s_addr;
	hp = gethostbyaddr((char *)&addr, sizeof(addr), AF_INET);
	if (hp != NULL)
		(void) sprintf(buf, "%s (%s)", hp->h_name, inet_ntoa(inaddr));
	else
		(void) sprintf(buf, "%s", inet_ntoa(inaddr));

	return(buf);
}


char *
pr_type(type, code)
u_char type;
u_char code;
{
	static char *ctab[] =
	{
	/*  0 */	"network is unreachable",
	/*  1 */	"host is unreachable",
	/*  2 */	"dest unreachable, protocol",
	/*  3 */	"dest unreachable, port",
	/*  4 */	"dest unreachable, needfrag",
	/*  5 */	"dest unreachable, srcfail"
	};

	static char *ttab[] =
	{
	/*  0 */	"echo reply",
	/*  1 */	"ICMP 1",
	/*  2 */	"ICMP 2",
	/*  3 */	"destination unreachable",
	/*  4 */	"source quench",
	/*  5 */	"redirect",
	/*  6 */	"ICMP 6",
	/*  7 */	"ICMP 7",
	/*  8 */	"echo request",
	/*  9 */	"ICMP 9",
	/* 10 */	"ICMP 10",
	/* 11 */	"time exceeded",
	/* 12 */	"parameter problem",
	/* 13 */	"timestamp request",
	/* 14 */	"timestamp reply",
	/* 15 */	"info request",
	/* 16 */	"info reply",
	/* 17 */	"mask request",
	/* 18 */	"mask reply"
	};

	if (type > 18)
		return("unknown packet type");

	if (type == ICMP_UNREACH)
		if (code <= 5)
			return(ctab[code]);

	return(ttab[type]);
}


u_short
checksum(addr, len)
u_short *addr;
int len;
{
	register int nleft = len;
	register u_short *w = addr;
	register int sum = 0;
	u_short answer = 0;

/*
 * Our algorithm is simple, using a 32 bit accumulator (sum),
 * we add sequential 16 bit words to it, and at the end, fold back
 * all the carry bits from the top 16 bits into the lower 16 bits.
 */
	while (nleft > 1)
	{
		sum += *w++;
		nleft -= 2;
	}

/*
 * Mop up an odd byte, if necessary.
 */
	if (nleft == 1)
	{
		*(u_char *)(&answer) = *(u_char *)w;
		sum += answer;
	}

/*
 * Add back carry outs from top 16 bits to low 16 bits.
 */
	sum = (sum >> 16) + (sum & 0xFFFF);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	answer = ~sum;				/* truncate to 16 bits */
	return(answer);
}


/*VARARGS1*/
error(fmt, a, b)
char *fmt;
{
	fprintf(stderr, "%s: ", program);
	fprintf(stderr, fmt, a, b);
	fprintf(stderr, "\n");
}

#ifdef OMNINET

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/fcntlcom.h>
#include <net/nit_if.h>
#include <sys/stropts.h>
#include <net/if.h>

int
initdevice(device)
	char *device;
{
	struct strioctl si;		/* struct for ioctl() */
	struct ifreq ifr;		/* interface request struct */
	u_long if_flags;		/* modes for interface */
	int  ret;
	int if_fd;

	if ((if_fd = open("/dev/nit", O_RDONLY)) < 0)
	{
		perror("/dev/nit");
		return(-1);
	}

	/* request the interface */
	strncpy(ifr.ifr_name, device, sizeof(ifr.ifr_name));
	ifr.ifr_name[sizeof(ifr.ifr_name) - 1] = ' ';
	si.ic_cmd = NIOCBIND;
	si.ic_len = sizeof(ifr);
	si.ic_dp = (char*)&ifr;
	if ((ret = ioctl(if_fd, I_STR, (char*)&si)) < 0)
	{
		perror(ifr.ifr_name);
		(void) close(if_fd);
		return(-1);
	}

	/* set the interface flags */
	si.ic_cmd = NIOCSFLAGS;
	if_flags = NI_PROMISC;
	si.ic_len = sizeof(if_flags);
	si.ic_dp = (char*)&if_flags;
	if ((ret = ioctl(if_fd, I_STR, (char*)&si)) < 0)
	{
		perror("NI_PROMISC");
		(void) close(if_fd);
		return(-1);
	}

	return(if_fd);
}

#endif OMNINET
